from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Dict, Final, List, Optional, Tuple, cast
import uuid

from google.protobuf.timestamp_pb2 import Timestamp
import grpc
import jsonpickle

from jetpack import __version__
from jetpack._remote import codec
from jetpack._runtime.local import LocalStub
from jetpack.config import k8s, namespace
from jetpack.proto.runtime.v1alpha1 import remote_pb2, remote_pb2_grpc

# Instead of having individual clients per functionality (e.g. job), have a
# single client for our runtime.

# TODO(Landau): Move client functionality out of _job.client and centralize it
# here.

# TODO(landau): make this a param
MAX_MESSAGE_LENGTH: Final[int] = 10485760  # 10MB

# Prevent circular dependency
if TYPE_CHECKING:
    from jetpack._task.task import Task


class JetpackException(Exception):
    """Base class for exceptions in this module"""

    pass


class RuntimeException(JetpackException):
    """Exception raised for errors in the Jetpack runtime and kubernetes."""

    def __init__(self, message: str) -> None:
        self.message = message


class ApplicationException(JetpackException):
    """Exception raised for errors from application-code that is using the SDK.

    TODO DEV-157
    For exceptions raised by remote functions and jobs, we serialize the
    userland exception in the backend and save it here. The userland exception
    is re-raised by the SDK for the caller of the remote function or job.
    """

    def __init__(self, message: str) -> None:
        self.message = message


class NoControllerAddressError(JetpackException):
    pass


class Client:
    def __init__(self) -> None:
        host = os.environ.get(
            "JETPACK_RUNTIME_SERVICE_HOST",
            "remotesvc.jetpack-runtime.svc.cluster.local",
        )
        port = os.environ.get("JETPACK_RUNTIME_SERVICE_PORT", "80")
        self.address: str = f"{host.strip()}:{port.strip()}"
        self.stub: Optional[remote_pb2_grpc.RemoteExecutorStub] = None

    def dial(self) -> remote_pb2_grpc.RemoteExecutorStub:
        if self.stub is None:
            if k8s.is_in_cluster():
                channel = grpc.aio.insecure_channel(
                    self.address,
                    options=[
                        ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
                    ],
                )
                self.stub = remote_pb2_grpc.RemoteExecutorStub(channel)
            else:
                self.stub = LocalStub(client)
        return self.stub

    async def set_cron_jobs(
        self,
        cron_jobs: List[remote_pb2.CronJob],
    ) -> remote_pb2.SetCronJobsResponse:
        request = remote_pb2.SetCronJobsRequest(
            namespace=cast(str, namespace.get()),
            hostname=os.environ["HOSTNAME"],
            cron_jobs=cron_jobs,
        )
        stub = self.dial()
        response: remote_pb2.SetCronJobsResponse = await stub.SetCronJobs(request)
        return response

    async def create_task(
        self,
        task: Task,
        args: Optional[Tuple[Any, ...]] = None,
        kwargs: Optional[Dict[str, Any]] = None,
    ) -> uuid.UUID:
        """Creates a Jetpack task. How the task gets executed is up to
        jetpack runtime.

        Keyword arguments:
        task -- task to create
        """

        request = self._build_create_task_request(task, args, kwargs)
        response: remote_pb2.CreateTaskResponse = await self.dial().CreateTask(request)
        return uuid.UUID(response.task_id)

    async def post_result(
        self, exec_id: str, value: Any = None, error: Optional[Exception] = None
    ) -> remote_pb2.PostResultResponse:

        if not exec_id:
            # Note: value=None and error=None is acceptable because a job
            # can run successfully and return nothing.
            raise Exception("An exec_id is required to post a result")

        result = remote_pb2.Result()
        if error:
            result.error.code = remote_pb2.APPLICATION
            result.error.message = str(error)
            result.error.encoded_error = bytes(jsonpickle.encode(error), "utf-8")
        else:  # order matters, as 'value' can be None
            result.value.encoded_value = bytes(jsonpickle.encode(value), "utf-8")

        request = remote_pb2.PostResultRequest(
            exec_id=exec_id,
            result=result,
        )

        response = await self.dial().PostResult(request)
        return cast(remote_pb2.PostResultResponse, response)

    async def wait_for_result(self, task_id: uuid.UUID) -> Any:
        request = remote_pb2.WaitForResultRequest(task_id=str(task_id))
        result = await self.dial().WaitForResult(request)
        return self._transform_response_exception(result)

    @staticmethod
    def _build_create_task_request(
        task: Task,
        args: Optional[Tuple[Any, ...]],
        kwargs: Optional[Dict[str, Any]],
    ) -> remote_pb2.CreateTaskRequest:
        encoded_args = b""
        if args or kwargs:
            encoded_args = codec.encode_args(
                args if args else None,
                kwargs if kwargs else None,
            ).encode("utf-8")

        current_namespace = namespace.get()
        task_proto_obj = remote_pb2.Task(
            qualified_symbol=task.jetpack_function.name(),
            encoded_args=encoded_args,
            hostname=os.environ.get("HOSTNAME", ""),  # k8s sets this
            target_time=Timestamp(seconds=task.target_time),
        )
        if current_namespace:
            task_proto_obj.namespace = current_namespace

        return remote_pb2.CreateTaskRequest(
            task=task_proto_obj, sdk_version=__version__
        )

    @staticmethod
    def _transform_response_exception(
        response: remote_pb2.WaitForResultResponse,
    ) -> Any:
        if response.result.HasField("error") and response.result.error.encoded_error:
            e = jsonpickle.decode(response.result.error.encoded_error)
            raise e
        elif response.result.HasField("value"):
            val = jsonpickle.decode(response.result.value.encoded_value)
            return val
        else:
            raise RuntimeException(
                f"Either 'value' or 'error' should be specified in response. Got: {response}"
            )


client = Client()



from concurrent import futures
from typing import Any, Callable, Dict

import grpc

from jetpack._remote import codec
from jetpack.proto.runtime.v1alpha1 import remote_pb2, remote_pb2_grpc


class Servicer(remote_pb2_grpc.RemoteExecutorServicer):
    def __init__(self) -> None:
        # TODO: Figure out if we need any locking around the symbol table.
        self.symbol_table: Dict[str, Callable[..., Any]] = {}  # TODO: lock?

    def export(self, fn: Callable[..., Any]) -> str:
        symbol = fn.__name__
        self.symbol_table[symbol] = fn
        return symbol

    def RemoteCall(
        self, request: remote_pb2.RemoteCallRequest, context: grpc.ServicerContext
    ) -> remote_pb2.RemoteCallResponse:
        fn = self.symbol_table[request.qualified_symbol]
        args, kwargs = codec.decode_args(request.json_args)
        result = fn(*args, **kwargs)
        return codec.encode_result(result)

    # Below are dummy implementation of LaunchJob(), LaunchBlockingJob(), and
    # PostResult() that throw NotImplemented exceptions.
    # These are required so that we fulfill the remote service interface.
    # In practice though, we should either unify the remote service
    # implementation in the SDK into a single class (this one) OR
    # separate the interface the SDK uses vs the interface the runtime uses.

    def CreateTask(
        self, request: remote_pb2.CreateTaskRequest, context: grpc.ServicerContext
    ) -> remote_pb2.CreateTaskResponse:
        return super().CreateTask(request, context)

    def PostResult(
        self,
        request: remote_pb2.PostResultRequest,
        context: grpc.ServicerContext,
    ) -> remote_pb2.PostResultResponse:
        return super().PostResult(request, context)

    def SetCronJobs(
        self,
        request: remote_pb2.SetCronJobsRequest,
        context: grpc.ServicerContext,
    ) -> remote_pb2.SetCronJobsResponse:
        return super().SetCronJobs(request, context)

    def WaitForResult(
        self,
        request: remote_pb2.WaitForResultRequest,
        context: grpc.ServicerContext,
    ) -> remote_pb2.WaitForResultResponse:
        return super().WaitForResult(request, context)


class Server:
    def __init__(self) -> None:
        self.servicer = Servicer()
        self.server = grpc.server(futures.ThreadPoolExecutor(max_workers=10))
        remote_pb2_grpc.add_RemoteExecutorServicer_to_server(self.servicer, self.server)
        self.is_listening = False  # TODO: Mutex needed?

    def Listen(self) -> None:
        self.server.add_insecure_port("[::]:50051")
        self.server.start()
        self.is_listening = True

    def export(self, fn: Callable[..., Any]) -> str:
        # Connect to the network lazily
        if not self.is_listening:
            self.Listen()
        return self.servicer.export(fn)



import asyncio
import time
from typing import Final

import click
import grpc

from jetpack import cron
from jetpack._runtime.client import client
from jetpack._task.jetpack_function import JetpackFunction
from jetpack.cmd import util
from jetpack.config import symbols
from jetpack.config.symbols import Symbol

ENTRYPOINT_PARAMS: Final = {
    "default": "/app",
    "envvar": "JETPACK_ENTRYPOINT",
    "help": "Path to user entrypoint that should be loaded. Can be file or directory. If directory, then we assume filename is main.py or jetpack_main.py.",
    "show_default": True,
}
_using_new_cli = False


def is_using_new_cli() -> bool:
    """
    for legacy uses, we keep old cli. This function disables that logic to ensure
    we don't run the cli command twice.
    """
    return _using_new_cli


@click.group()
def cli() -> None:
    global _using_new_cli
    _using_new_cli = True


@click.command(help="Executes jetpack task")
@click.option("--entrypoint", **ENTRYPOINT_PARAMS)
@click.option("--exec-id", required=True)
@click.option("--qualified-symbol", required=True)
@click.option("--encoded-args", default="")
def exec_task(
    entrypoint: str,
    exec_id: str,
    qualified_symbol: str,
    encoded_args: str,
) -> None:
    util.load_user_entrypoint(entrypoint)
    func = symbols.get_symbol_table()[Symbol(qualified_symbol)]
    asyncio.run(JetpackFunction(func).exec(exec_id, encoded_args))


@click.command(help="Executes cronjob")
@click.option("--entrypoint", **ENTRYPOINT_PARAMS)
@click.option("--qualified-symbol", required=True)
def exec_cronjob(entrypoint: str, qualified_symbol: str) -> None:
    util.load_user_entrypoint(entrypoint)
    func = symbols.get_symbol_table()[Symbol(qualified_symbol)]
    asyncio.run(JetpackFunction(func).exec(post_result=False))


@click.command(help="Registers jetpack functions with runtime")
@click.option("--entrypoint", **ENTRYPOINT_PARAMS)
def register(entrypoint: str) -> None:
    async def _register() -> None:
        util.load_user_entrypoint(entrypoint)
        # For now, we try a few times to register with runtime. Once the runtime
        # becomes a sidecar, we can remove this.
        tries = 3
        for i in range(tries):
            try:
                await client.set_cron_jobs(cron.get_jobs())
            except grpc.RpcError as rpc_error:
                if rpc_error.code() == grpc.StatusCode.UNAVAILABLE:
                    if i == tries - 1:
                        # no more tries, just give up.
                        raise rpc_error
                    print("runtime is not available. Sleep(5) and try again")
                    time.sleep(5)
                else:
                    raise rpc_error

    asyncio.run(_register())


@click.command(help="List jetroutines")
@click.option("--entrypoint", **ENTRYPOINT_PARAMS)
def ls(entrypoint: str) -> None:
    util.load_user_entrypoint(entrypoint)
    print(symbols.get_symbol_table().defined_symbols())


cli.add_command(exec_task)
cli.add_command(exec_cronjob)
cli.add_command(register)
cli.add_command(ls)
